@@ -2,15 +2,15 @@
 GULP Metal Optimizer
 ==============
 
-#. :class:`.GulpMetalOptimizer`
-#. :class:`.GulpMDMetalOpimizer`
+#. :class:`.GulpOptimizer`
+#. :class:`.GulpMDOpimizer`
 
 Wrappers for calculators within the :mod:`gulp` code.
 
 Examples
 --------
 
-While metal atoms are not required, UFF4MOF is a useful because it
+While metal atoms are not required, UFF4MOF is useful because it
 encompasses almost all chemical environments commonly found in
 metal-organic structures. Better forcefields exist for purely
 organic molecules! An interface with GULP is provided, which takes
@@ -74,7 +74,7 @@ optimisations.
     # Use conjugate gradient method for a slower, but more stable
     # optimisation.
 
-    gulp_opt = stko.GulpMetalOptimizer(
+    gulp_opt = stko.GulpUFFOptimizer(
         gulp_path='path/to/gulp',
         metal_FF={46: 'Pd4+2'},
         conjugate_gradient=True
@@ -95,7 +95,7 @@ or a more robust method!
 
 .. code-block:: python
 
-    gulp_MD = stko.GulpMDMetalOptimizer(
+    gulp_MD = stko.GulpUFFMDOptimizer(
         gulp_path='path/to/gulp',
         metal_FF={46: 'Pd4+2'},
         temperature=700,
@@ -128,11 +128,19 @@ from ..utilities import (
 logger = logging.getLogger(__name__)
 
 
+class ExpectedMetal(Exception):
+    ...
+
+
+class ExpectedCell(Exception):
+    ...
+
+
 class UFFTyperError(Exception):
     ...
 
 
-class GulpMetalOptimizer(Optimizer):
+class GulpUFFOptimizer(Optimizer):
     """
     Applies forcefield optimizers that can handle metal centres.
 
@@ -147,36 +155,60 @@ class GulpMetalOptimizer(Optimizer):
     def __init__(
         self,
         gulp_path,
-        metal_FF,
+        maxcyc=1000,
+        metal_FF=None,
+        metal_ligand_bond_order=None,
         conjugate_gradient=False,
+        periodic=False,
         output_dir=None,
     ):
         """
-        Initialize a :class:`GulpMetalOptimizer` instance.
+        Initialize a :class:`GulpUFFOptimizer` instance.
 
         Parameters
         ----------
         gulp_path : :class:`str`
             Path to GULP executable.
 
-        metal_FF : :class:`dict`
+        maxcyc : :class:`int`
+            Set the maximum number of optimisation steps to use.
+            Default in Gulp is 1000.
+
+        metal_FF : :class:`dict`, optional
             Dictionary with metal atom forcefield assignments.
             Key: :class:`int` : atomic number.
             Value: :class:`str` : UFF4MOF forcefield type.
 
-        conjugate_gradient : :class:``, optional
+        metal_ligand_bond_order : :class:`str`, optional
+            Bond order to use for metal-ligand bonds. Defaults to
+            `half`, but using `resonant` can increase the force
+            constant for stronger metal-ligand interactions.
+
+        conjugate_gradient : :class:`bool`, optional
             ``True`` to use Conjugate Graditent method.
             Defaults to ``False``
 
+        periodic : :class:`bool`, optional
+            If `False`, then calculation is non-periodic.
+            Periodic optimisations always optimize the cell parameters
+            at constant pressure.
+
         output_dir : :class:`str`, optional
             The name of the directory into which files generated during
             the calculation are written, if ``None`` then
             :func:`uuid.uuid4` is used.
 
         """
+
         self._gulp_path = gulp_path
+        self._maxcyc = maxcyc
         self._metal_FF = metal_FF
+        self._metal_ligand_bond_order = (
+            'half' if metal_ligand_bond_order is None
+            else metal_ligand_bond_order
+        )
         self._conjugate_gradient = conjugate_gradient
+        self._periodic = periodic
         self._output_dir = output_dir
 
     def _add_atom_charge_flags(self, atom, atomkey):
@@ -462,6 +494,21 @@ class GulpMetalOptimizer(Optimizer):
 
         return type_translator
 
+    def _cell_section(self, cell):
+        cell_section = (
+            '\ncell\n'
+            f"{round(cell.a, 6)} "
+            f"{round(cell.b, 6)} "
+            f"{round(cell.c, 6)} "
+            f"{round(cell.alpha, 6)} "
+            f"{round(cell.beta, 6)} "
+            f"{round(cell.gamma, 6)} "
+            # No fixes.
+            "0 0 0 0 0 0\n"
+        )
+
+        return cell_section
+
     def _position_section(self, mol, type_translator):
         position_section = '\ncartesian\n'
         for atom in mol.get_atoms():
@@ -490,7 +537,7 @@ class GulpMetalOptimizer(Optimizer):
                 # H has bond order of 1.
                 bond_type = ''
             elif has_metal_atom(bond, metal_atoms):
-                bond_type = 'half'
+                bond_type = self._metal_ligand_bond_order
             elif '_R' in atom_types[0] and '_R' in atom_types[1]:
                 bond_type = 'resonant'
             elif bond.get_order() == 1:
@@ -516,16 +563,36 @@ class GulpMetalOptimizer(Optimizer):
 
         return species_section
 
-    def _write_gulp_file(self, mol, metal_atoms, in_file, output_xyz):
+    def _write_gulp_file(
+        self,
+        mol,
+        metal_atoms,
+        in_file,
+        output_xyz,
+        cell=None,
+    ):
 
         type_translator = self._type_translator()
 
+        top_line = 'opti '
+
         if self._conjugate_gradient:
-            top_line = (
-                'opti conj unit conv noautobond fix molmec cartesian\n'
-            )
+            top_line += 'conj unit '
+
+        if self._periodic:
+            # Constant pressure.
+            top_line += 'conp '
+            cell_section = self._cell_section(cell)
+            # Output CIF.
+            output_cif = output_xyz.replace('xyz', 'cif')
+            periodic_output = f'output cif {output_cif}\n'
         else:
-            top_line = 'opti conv noautobond fix molmec cartesian\n'
+            # Constant volume.
+            top_line += 'conv '
+            cell_section = ''
+            periodic_output = ''
+
+        top_line += 'noautobond fix molmec cartesian\n'
 
         position_section = self._position_section(mol, type_translator)
         bond_section = self._bond_section(mol, metal_atoms)
@@ -535,16 +602,19 @@ class GulpMetalOptimizer(Optimizer):
 
         output_section = (
             '\n'
+            f'maxcyc {self._maxcyc}\n'
             'terse inout potentials\n'
-            'terse inout cell\n'
+            'terse in cell\n'
             'terse in structure\n'
             'terse inout derivatives\n'
             f'output xyz {output_xyz}\n'
+            f'{periodic_output}'
             # 'output movie xyz steps_.xyz\n'
         )
 
         with open(in_file, 'w') as f:
             f.write(top_line)
+            f.write(cell_section)
             f.write(position_section)
             f.write(bond_section)
             f.write(species_section)
@@ -557,15 +627,23 @@ class GulpMetalOptimizer(Optimizer):
 
         Parameters
         ----------
+        mol : :class:`stk.Molecule`
+            The molecule to be optimized.
 
         Returns
         -------
-
         None : :class:`NoneType`
 
         """
         metal_atoms = get_metal_atoms(mol)
         metal_ids = [i.get_id() for i in metal_atoms]
+
+        if len(metal_ids) > 1 and self._metal_FF is None:
+            raise ExpectedMetal(
+                'No metal FF provivded, but metal atoms were found ('
+                f'{metal_atoms})'
+            )
+
         metal_bonds, _ = get_metal_bonds(mol, metal_atoms)
         edit_mol = to_rdkit_mol_without_metals(
             mol=mol,
@@ -607,14 +685,6 @@ class GulpMetalOptimizer(Optimizer):
                 shell=True
             )
 
-    def _move_generated_files(self, files):
-        if not os.path.exists(self._output_dir):
-            os.mkdir(self._output_dir)
-
-        for file in files:
-            if os.path.exists(file):
-                os.rename(file, f'{self._output_dir}/{file}')
-
     def extract_final_energy(self, file):
         nums = re.compile(r"[+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?")
         with open(file, 'r') as f:
@@ -623,21 +693,30 @@ class GulpMetalOptimizer(Optimizer):
                     string = nums.search(line.rstrip()).group(0)
                     return float(string)
 
-    def optimize(self, mol):
+    def optimize(self, mol, cell=None, cif_filename=None):
         """
-        Optimize `mol`.
+        Optimize `mol` and `cell`.
 
         Parameters
         ----------
-        mol : :class:`.Molecule`
+        mol : :class:`stk.Molecule`
             The molecule to be optimized.
 
+        cell : :class:`.Cell`, optional
+            The cell to be optimized if optimization is periodic.
+
         Returns
         -------
         mol : :class:`.Molecule`
             The optimized molecule.
 
         """
+
+        if cell is None and self._periodic:
+            raise ExpectedCell(
+                'Optimisation expected a cell because periodic is True'
+            )
+
         if self._output_dir is None:
             output_dir = str(uuid.uuid4().int)
         else:
@@ -648,6 +727,8 @@ class GulpMetalOptimizer(Optimizer):
             shutil.rmtree(output_dir)
 
         os.mkdir(output_dir)
+        init_dir = os.getcwd()
+        os.chdir(output_dir)
 
         in_file = 'gulp_opt.gin'
         out_file = 'gulp_opt.ginout'
@@ -655,29 +736,28 @@ class GulpMetalOptimizer(Optimizer):
 
         metal_atoms = get_metal_atoms(mol)
 
-        # Write GULP file.
-        self._write_gulp_file(
-            mol=mol,
-            metal_atoms=metal_atoms,
-            in_file=in_file,
-            output_xyz=output_xyz
-        )
+        try:
+            # Write GULP file.
+            self._write_gulp_file(
+                mol=mol,
+                metal_atoms=metal_atoms,
+                in_file=in_file,
+                output_xyz=output_xyz,
+                cell=cell,
+            )
+            # Run.
+            self._run_gulp(in_file, out_file)
 
-        # Run.
-        self._run_gulp(in_file, out_file)
+            # Update from output.
+            mol = mol.with_structure_from_file(output_xyz)
 
-        # Update from output.
-        mol = mol.with_structure_from_file(output_xyz)
-
-        # Move files.
-        self._move_generated_files(
-            files=[in_file, out_file, output_xyz]
-        )
+        finally:
+            os.chdir(init_dir)
 
         return mol
 
 
-class GulpMDMetalOptimizer(GulpMetalOptimizer):
+class GulpUFFMDOptimizer(GulpUFFOptimizer):
     """
     Applies forcefield MD that can handle metal centres.
 
@@ -692,7 +772,8 @@ class GulpMDMetalOptimizer(GulpMetalOptimizer):
     def __init__(
         self,
         gulp_path,
-        metal_FF,
+        metal_FF=None,
+        metal_ligand_bond_order=None,
         output_dir=None,
         integrator='stochastic',
         ensemble='nvt',
@@ -705,18 +786,23 @@ class GulpMDMetalOptimizer(GulpMetalOptimizer):
         save_conformers=False,
     ):
         """
-        Initialize a :class:`GulpMetalOptimizer` instance.
+        Initialize a :class:`GulpUFFOptimizer` instance.
 
         Parameters
         ----------
         gulp_path : :class:`str`
             Path to GULP executable.
 
-        metal_FF : :class:`dict`
+        metal_FF : :class:`dict`, optional
             Dictionary with metal atom forcefield assignments.
             Key: :class:`int` : atomic number.
             Value: :class:`str` : UFF4MOF forcefield type.
 
+        metal_ligand_bond_order : :class:`str`, optional
+            Bond order to use for metal-ligand bonds. Defaults to
+            `half`, but using `resonant` can increase the force
+            constant for stronger metal-ligand interactions.
+
         output_dir : :class:`str`, optional
             The name of the directory into which files generated during
             the calculation are written, if ``None`` then
@@ -761,6 +847,10 @@ class GulpMDMetalOptimizer(GulpMetalOptimizer):
         """
         self._gulp_path = gulp_path
         self._metal_FF = metal_FF
+        self._metal_ligand_bond_order = (
+            'half' if metal_ligand_bond_order is None
+            else metal_ligand_bond_order
+        )
         self._output_dir = output_dir
         self._integrator = integrator
         self._ensemble = ensemble
@@ -838,105 +928,102 @@ class GulpMDMetalOptimizer(GulpMetalOptimizer):
             lines = f.readlines()
 
         # Split file using strings.
-        id = 0
-        s_times = {}
-        s_coords = {}
-        s_vels = {}
-        s_derivs = {}
-        s_sites = {}
-        coords = []
-        vels = []
-        derivs = []
-        sites = []
-        switch = None
-        for line in lines:
-            li = line.rstrip()
-            if '#  Time/KE/E/T' in li:
-                switch = 'times'
-                s_coords[id] = coords
-                coords = []
-                s_vels[id] = vels
-                vels = []
-                s_derivs[id] = derivs
-                derivs = []
-                s_sites[id] = sites
-                sites = []
-                id += 1
-            elif '#  Coordinates' in li:
-                switch = 'coords'
-            elif '#  Velocities' in li:
-                switch = 'vels'
-            elif '#  Derivatives' in li:
-                switch = 'derivs'
-            elif '#  Site energies' in li:
-                switch = 'sites'
-            elif switch == 'coords':
-                coords.append([i for i in li.split(' ') if i])
-            elif switch == 'vels':
-                vels.append([i for i in li.split(' ') if i])
-            elif switch == 'derivs':
-                derivs.append([i for i in li.split(' ') if i])
-            elif switch == 'sites':
-                sites.append([i for i in li.split(' ') if i])
-            elif switch == 'times':
-                s_times[id] = [i for i in li.split(' ') if i]
-            elif switch is None:
-                pass
-        # Add final timestep.
-        s_coords[id] = coords
-        coords = []
-        s_vels[id] = vels
-        vels = []
-        s_derivs[id] = derivs
-        derivs = []
-        s_sites[id] = sites
-        sites = []
-
-        ids = []
-        tt = []
-        pot_energies = []
-        new_lines = []
-        for id in s_times:
-            times = s_times[id]
-            ids.append(id)
-            tt.append(float(times[0]))
-            pot_energies.append(float(times[2]))
-            coords = s_coords[id]
-            sites = s_sites[id]
-            xyz_string = (
-                f'{len(coords)}\n'
-                f'{times[0]},{times[1]},{times[2]},{times[3]}\n'
+        timesteps = ''.join(lines).split('#  Time/KE/E/T')[1:]
+        trajectory_data = {}
+        xyz_traj_lines = []
+        for ts, cont in enumerate(timesteps):
+            ts_data = {}
+            time_section = (
+                cont.split('#  Coordinates\n')[0]
+            )
+            coords_section = (
+                cont.split('#  Coordinates\n')[1].split(
+                    '#  Velocities\n'
+                )[0]
+            )
+            vels_section = (
+                cont.split('#  Velocities\n')[1].split(
+                    '#  Derivatives \n'
+                )[0]
+            )
+            derivs_section = (
+                cont.split('#  Derivatives \n')[1].split(
+                    '#  Site energies \n'
+                )[0]
+            )
+            sites_section = (
+                cont.split('#  Site energies \n')[1]
             )
 
-            for i, coord in enumerate(coords):
-                site = sites[i][0]
+            ts_data['time'] = float(
+                [i for i in time_section.strip().split(' ') if i][0]
+            )
+            ts_data['KE'] = float(
+                [i for i in time_section.strip().split(' ') if i][1]
+            )
+            ts_data['E'] = float(
+                [i for i in time_section.strip().split(' ') if i][2]
+            )
+            ts_data['T'] = float(
+                [i for i in time_section.strip().split(' ') if i][3]
+            )
+            ts_data['coords'] = [
+                [i for i in li.split(' ') if i]
+                for li in coords_section.split('\n')[:-1]
+            ]
+            ts_data['vels'] = [
+                [i for i in li.split(' ') if i]
+                for li in vels_section.split('\n')[:-1]
+            ]
+            ts_data['derivs'] = [
+                [i for i in li.split(' ') if i]
+                for li in derivs_section.split('\n')[:-1]
+            ]
+            ts_data['sites'] = [
+                [i for i in li.split(' ') if i]
+                for li in sites_section.split('\n')[:-1]
+            ]
+
+            trajectory_data[ts] = ts_data
+
+            # Write XYZ string for XYZ traj file.
+            xyz_string = (
+                f"{len(ts_data['coords'])}\n"
+                f"{ts_data['time']},{ts_data['KE']},"
+                f"{ts_data['E']},{ts_data['T']}\n"
+            )
+
+            for i, coord in enumerate(ts_data['coords']):
+                site_E = ts_data['sites'][i][0]
                 xyz_string += (
                     f'{atom_types[i]} {coord[0]} {coord[1]} '
-                    f'{coord[2]} {site}\n'
+                    f'{coord[2]} {site_E}\n'
                 )
 
-            new_lines.append(xyz_string)
+            xyz_traj_lines.append(xyz_string)
 
+        # Write XYZ trajectory file.
         with open(xyz_traj, 'w') as f:
-            for line in new_lines:
+            for line in xyz_traj_lines:
                 f.write(line)
 
-        return atom_types, ids, tt, pot_energies, s_times, s_coords
+        return atom_types, trajectory_data
 
     def _write_conformer_xyz_file(
         self,
-        id,
+        ts,
+        ts_data,
         filename,
-        s_times,
-        s_coords,
-        atom_types
+        atom_types,
     ):
-        times = s_times[id]
-        coords = s_coords[id]
+
+        coords = ts_data['coords']
         xyz_string = (
-            f'{len(coords)}\n'
-            f'{times[0]},{times[1]},{times[2]},{times[3]}\n'
+            f"{len(ts_data['coords'])}\n"
+            f"{ts},{ts_data['time']},{ts_data['KE']},"
+            f"{ts_data['E']},{ts_data['T']}\n"
         )
+
         for i, coord in enumerate(coords):
             xyz_string += (
                 f'{atom_types[i]} {coord[0]} {coord[1]} {coord[2]}\n'
@@ -954,82 +1041,74 @@ class GulpMDMetalOptimizer(GulpMetalOptimizer):
     ):
 
         # Convert GULP trajectory file to xyz trajectory.
-        results = self._convert_traj_to_xyz(
+        atom_types, trajectory_data = self._convert_traj_to_xyz(
             output_xyz,
             output_traj,
             xyz_traj
         )
-        atom_types, ids, tt, pot_energies, s_times, s_coords = results
 
         # Find lowest energy conformation and output to XYZ.
         if self._opt_conformers:
             # Optimise all conformers.
             min_energy = 1E10
-            for id in ids:
+            for ts in trajectory_data:
                 if self._save_conformers:
-                    conformer_file_name = os.path.join(
-                        self._output_dir,
-                        f'conf_{id}.xyz'
-                    )
+                    conformer_file_name = f'conf_{ts}.xyz'
                 else:
                     # This will get overwrriten each time.
-                    conformer_file_name = os.path.join(
-                        self._output_dir,
-                        'temp_conf.xyz'
-                    )
+                    conformer_file_name = 'temp_conf.xyz'
+
                 self._write_conformer_xyz_file(
-                    id=id,
+                    ts=ts,
+                    ts_data=trajectory_data[ts],
                     filename=conformer_file_name,
-                    s_times=s_times,
-                    s_coords=s_coords,
                     atom_types=atom_types
                 )
                 mol = mol.with_structure_from_file(conformer_file_name)
-                gulp_opt = GulpMetalOptimizer(
+                conformer_opt_dir = os.path.join(
+                    os.getcwd(),
+                    f'conf_{ts}_opt'
+                )
+                gulp_opt = GulpUFFOptimizer(
                     gulp_path=self._gulp_path,
                     metal_FF=self._metal_FF,
-                    output_dir=self._output_dir
+                    output_dir=conformer_opt_dir,
                 )
                 gulp_opt.assign_FF(mol)
                 mol = gulp_opt.optimize(mol=mol)
                 energy = gulp_opt.extract_final_energy(
                     file=os.path.join(
-                        self._output_dir,
+                        conformer_opt_dir,
                         'gulp_opt.ginout'
                     )
                 )
-
                 if energy < min_energy:
                     min_energy = energy
-                    self._write_conformer_xyz_file(
-                        id=id,
-                        filename=low_conf_xyz,
-                        s_times=s_times,
-                        s_coords=s_coords,
-                        atom_types=atom_types
-                    )
+                    # Write out optimised conformer.
+                    mol.write(low_conf_xyz)
         else:
             if self._save_conformers:
-                for id in ids:
-                    conformer_file_name = os.path.join(
-                        self._output_dir,
-                        f'conf_{id}.xyz'
-                    )
+                for ts in trajectory_data:
+                    conformer_file_name = f'conf_{ts}.xyz'
                     self._write_conformer_xyz_file(
-                        id=id,
+                        ts=ts,
+                        ts_data=trajectory_data[ts],
                         filename=conformer_file_name,
-                        s_times=s_times,
-                        s_coords=s_coords,
                         atom_types=atom_types
                     )
 
-            min_energy = min(pot_energies)
-            min_id = ids[pot_energies.index(min_energy)]
+            energies = [
+                trajectory_data[ts]['E']
+                for ts in trajectory_data
+            ]
+            min_energy = min(energies)
+            min_ts = list(trajectory_data.keys())[
+                energies.index(min_energy)
+            ]
             self._write_conformer_xyz_file(
-                id=min_id,
+                ts=min_ts,
+                ts_data=trajectory_data[min_ts],
                 filename=low_conf_xyz,
-                s_times=s_times,
-                s_coords=s_coords,
                 atom_types=atom_types
             )
 
@@ -1039,15 +1118,16 @@ class GulpMDMetalOptimizer(GulpMetalOptimizer):
 
         Parameters
         ----------
-        mol : :class:`.Molecule`
+        mol : :class:`stk.Molecule`
             The molecule to be optimized.
 
         Returns
         -------
-        mol : :class:`.Molecule`
+        mol : :class:`stk.Molecule`
             The optimized molecule.
 
         """
+
         if self._output_dir is None:
             output_dir = str(uuid.uuid4().int)
         else:
@@ -1058,47 +1138,44 @@ class GulpMDMetalOptimizer(GulpMetalOptimizer):
             shutil.rmtree(output_dir)
 
         os.mkdir(output_dir)
+        init_dir = os.getcwd()
+        os.chdir(output_dir)
 
         in_file = 'gulp_MD.gin'
         out_file = 'gulp_MD.ginout'
         output_xyz = 'gulp_MD_template.xyz'
-        mol.write(output_xyz)
         output_traj = 'gulp_MD.trg'
         xyz_traj = 'gulp_MD_traj.xyz'
         low_conf_xyz = 'low_energy_conf.xyz'
 
         metal_atoms = get_metal_atoms(mol)
 
-        # Write GULP file.
-        self._write_gulp_file(
-            mol=mol,
-            metal_atoms=metal_atoms,
-            in_file=in_file,
-            output_traj=output_traj
-        )
+        try:
+            mol.write(output_xyz)
+            # Write GULP file.
+            self._write_gulp_file(
+                mol=mol,
+                metal_atoms=metal_atoms,
+                in_file=in_file,
+                output_traj=output_traj
+            )
 
-        # Run.
-        self._run_gulp(in_file, out_file)
+            # Run.
+            self._run_gulp(in_file, out_file)
 
-        # Get lowest energy conformer from trajectory.
-        self._save_lowest_energy_conf(
-            mol=mol,
-            output_xyz=output_xyz,
-            output_traj=output_traj,
-            xyz_traj=xyz_traj,
-            low_conf_xyz=low_conf_xyz
-        )
+            # Get lowest energy conformer from trajectory.
+            self._save_lowest_energy_conf(
+                mol=mol,
+                output_xyz=output_xyz,
+                output_traj=output_traj,
+                xyz_traj=xyz_traj,
+                low_conf_xyz=low_conf_xyz
+            )
 
-        # Update from output.
-        mol = mol.with_structure_from_file(low_conf_xyz)
+            # Update from output.
+            mol = mol.with_structure_from_file(low_conf_xyz)
 
-        # Move files.
-        self._move_generated_files(
-            files=[
-                in_file, out_file, output_xyz,
-                output_traj, xyz_traj, low_conf_xyz,
-                'temp_conf.xyz'
-            ]
-        )
+        finally:
+            os.chdir(init_dir)
 
         return mol
